module emop.top {
   main :-
      new.(
         import(emop.syntax:syntax),
         loop
      ).

   new(Top) :-
      Scope = data.ref.backtrackable:new([]),
      ParentVars = data.record:{},
      SgnVars = data.record:{},
      Prompt = data.ref.backtrackable:new('(~d) ?- '),
      module Top {
         import: module [Prompt] {
            import: emop.sgn.
            
            read_line(Acc, Ast) :-
               Line = data.atom:concat(Acc, io.std:read_line),
               (
                  Line = ''
               ->
                  throw('empty_line')
               ;
                  Line = 'backtrack'
               ->
                  throw('backtrack')
               ;
                  true
               ),
               (
                  Ast = emop.parser:parse_stream(
                     '', 'toplevel', data.atom:stream(Line), emop.sgn
                  )
               ->
                  true
               ;
                  read_line(Line, Ast)
               ).
            
            read_line = read_line('').
            
            update = module {
               update(Filename) :-
                  io.std:format('The file \'~a\' has changed.  Reload? (y/n) ', [Filename]),
                  repeat,
                  Answer = io.std:read_line,
                  (
                     Answer = 'y'
                  ->
                     !,
                     true
                  ;
                     Answer = 'n'
                  ->
                     !,
                     fail
                  ;
                     io.std:print('Please answer "y" or "n". '),
                     fail
                  ).
            }.
            
            prompt_goal(Level, ParentVars, SgnVars, Scope, NewScope, Vars, M) :-
               print_store(ParentVars),
               io.std:format(Prompt:get, [Level]),
               Ast = read_line,
               (
                  catch(
                     (
                        Directive = emop.sgn:directive_syntax(Ast)
                     ->
                        execute_directive(Directive, ParentVars, SgnVars, Scope, NewScope, Vars, M)
                     ;
                        emop.compile:load_ast(Ast, update, ParentVars:keys, SgnVars, Scope, Vars, M),
                        NewScope = Scope
                     ),
                     E,
                     (
                        emop.message:error(E),
                        prompt_goal(Level, ParentVars, SgnVars, Scope, NewScope, Vars, M)
                     )
                  )
               ->
                  true
               ;
                  io.err:print_endline('top-level bug'),
                  fail
               ).
            
            execute_directive(Directive, ParentVars, SgnVars, Scope, NewScope, Vars, M) :-
               (
                  Directive = 'special'('import', S, Pb, Pe)
               ->
                  emop.compile:import(S, Pb, Pe, update, ParentVars:keys, SgnVars, Scope, NewScope, Vars, M)
               ;
                  Directive = 'special'('abstract', _, Pb, Pe)
               ->
                  throw(error('Directive abstract cannot be used outside a module', file(Pb, Pe)))
               ;
                  directive_sgn(Directive, [], Sgn),
                  NewScope = [emop.expand:make_scope('', Sgn, Scope) | Scope],
                  M = module { do(_). },
                  Vars = ParentVars
               ).
            
            print_store(Vars) :-
               (
                  Names = data.list:merge_sort(Vars:keys, data.term),
                  data.list:iter(data.list:reverse(Names)) {
                     do(Var) :-
                        Vars:{Var} = 'var'(Value),
                        (
                           Value = Var
                        ->
                           true
                        ;
                           true
                        )
                  },
                  data.list:iter(Names) {
                     do(Var) :-
                        Vars:{Var} = 'var'(Value),
                        (
                           (
                              Value = Var
                           ;
                              data.atom:sub(Var, 0, 1, _, '_')
                           )
                        ->
                           true
                        ;
                           io.std:format('~a = ~p\n', [Var, Value])
                        ).
                  },
                  fail
               ;
                  true
               ).
            
            loop(Level, ParentVars, SgnVars, Scope) :-
               (
                  repeat,
                  prompt_and_execute(Level, ParentVars, SgnVars, Scope)
               ->
                  true
               ).
            
            init_vars(Vars, NewVars) = data.list:map(Vars) {
               do(Var, Value) :-
                  NewVars:{Var} = 'var'(Value).
            }.
            
            prompt_and_execute(Level, ParentVars, SgnVars, Scope) :-
               catch(
                  M = prompt_goal(Level, ParentVars, SgnVars, Scope, NewScope, Vars),
                  'empty_line',
                  (
                     (
                        data.number:(Level > 0)
                     ->
                        throw('other')
                     ;
                        io.std:print_endline('No goal to enumerate.'),
                        fail
                     )
                  )
               ),
               catch(
                  catch(
                     (
                        NewVars = ParentVars:copy,
                        init_vars(Vars, NewVars, Values),
                        call_det(M:do(data.term:make('vars', Values)), Det),
                        io.std:endline,
                        (
                           Det = 'true'
                        ->
                           true
                        ;
                           io.std:print_endline('There may be more solutions.')
                        ),
                        loop_succ(Det, data.number:succ(Level), NewVars, SgnVars, NewScope)
                     ;
                        io.err:print_endline('no'),
                        fail
                     ),
                     'backtrack',
                     (
                        !,
                        fail
                     )
                  ),
                  E,
                  (
                     (
                        E = 'end_of_file'
                     ->
                        throw('end_of_file')
                     ;
                        io.err:format('Uncaught exception: ~q\n', [E]),
                        loop(Level, ParentVars, SgnVars, Scope)
                     )
                  )
               ).
            
            loop_succ(Det, Level, NewVars, SgnVars, Scope) :-
               catch(
                  loop(Level, NewVars, SgnVars, Scope),
                  'other',
                  (
                     Det = 'true'
                  ->
                     io.std:print_endline('There is no other solution.'),
                     loop_succ(Det, Level, NewVars, SgnVars, Scope)
                  ;
                     fail
                  )
               ).
         }.

         import(M) :-
            SgnList = control.iter:findall() {
               do = M:`$sgn'.
            },
            Sgn = data.list:flatten(SgnList),
            Var = emop.parser:fresh_var,
            Var = 'var'(V),
            ParentVars:{V} = M,
            SgnVars:{V} = Sgn,
            Old = Scope:get,
            Scope:set([emop.expand:make_scope(Var, Sgn, Old) | Old]).

         prompt = Prompt.

         loop :-
            catch(
               catch(
                  loop(0, ParentVars, SgnVars, Scope:get),
                  'backtrack',
                  true
               ),
               'end_of_file',
               true
            ).
      }.
}